import logging

from ai.backend.common.identity import get_instance_id, get_instance_ip

log = logging.getLogger('ai.backend.gateway.etcd')


class ConfigServer:

    def __init__(self, config):
        # WARNING: importing etcd3/grpc must be done after forks.
        self.config = config
        from ai.backend.common.etcd import AsyncEtcd
        self.etcd = AsyncEtcd(config.etcd_addr, config.namespace)

    async def register_myself(self):
        instance_id = await get_instance_id()
        instance_ip = await get_instance_ip()
        event_addr = f'{instance_ip}:{self.config.events_port}'
        await self.etcd.put_multi(
            ['nodes/manager', 'nodes/redis', 'nodes/manager/event_addr'],
            [instance_id, self.config.redis_addr, event_addr])

    async def deregister_myself(self):
        await self.etcd.delete_prefix('nodes/manager')

    async def update_kernel_images(self):
        pass
        # TODO: a method to scan docker hub and update kernel image versions
        # TODO: a cli command to execute the above method


async def init(app):
    app['config_server'] = ConfigServer(app.config)
    await app['config_server'].register_myself()


async def shutdown(app):
    await app['config_server'].deregister_myself()



from decimal import Decimal
import logging
import time

import aioredis

from .defs import REDIS_RLIM_DB
from .exceptions import RateLimitExceeded

log = logging.getLogger('ai.backend.gateway.ratelimit')

_time_prec = Decimal('1e-3')  # msec
_rlim_window = 60 * 15        # 15 minutes


async def rlim_middleware_factory(app, handler):
    async def rlim_middleware_handler(request):
        # TODO: use a global timer if we scale out the gateway.
        now = Decimal(time.monotonic()).quantize(_time_prec)
        if request['is_authorized']:
            rate_limit = request['keypair']['rate_limit']
            access_key = request['keypair']['access_key']
            rlim_next_reset = now + _rlim_window
            async with app['redis_rlim'].get() as rr:
                tracker = await rr.hgetall(access_key)
                if tracker:
                    rlim_reset = Decimal(tracker['rlim_reset'])
                if not tracker or rlim_reset <= now:
                    # create a new tracker info
                    tracker = {
                        'rlim_limit': rate_limit,
                        'rlim_remaining': rate_limit,
                        'rlim_reset': str(rlim_next_reset),
                    }
                else:
                    tracker = {
                        'rlim_limit': rate_limit,
                        'rlim_remaining': int(tracker['rlim_remaining']) - 1,
                        'rlim_reset': tracker['rlim_reset'],  # copy
                    }
                await rr.hmset_dict(access_key, tracker)
            if tracker['rlim_remaining'] <= 0:
                raise RateLimitExceeded
            rlim_reset = Decimal(tracker['rlim_reset'])
            count_remaining = tracker['rlim_remaining']
            window_remaining = int((rlim_reset - now) * 1000)
            response = await handler(request)
            response.headers['X-RateLimit-Limit'] = str(rate_limit)
            response.headers['X-RateLimit-Remaining'] = str(count_remaining)
            response.headers['X-RateLimit-Reset'] = str(window_remaining)
            return response
        else:
            # No checks for rate limiting for non-authorized queries.
            response = await handler(request)
            response.headers['X-RateLimit-Limit'] = '1000'
            response.headers['X-RateLimit-Remaining'] = '1000'
            response.headers['X-RateLimit-Reset'] = str(_rlim_window * 1000)
            return response
    return rlim_middleware_handler


async def init(app):
    app['redis_rlim'] = await aioredis.create_pool(
        app.config.redis_addr.as_sockaddr(),
        create_connection_timeout=3.0,
        encoding='utf8',
        db=REDIS_RLIM_DB)
    app.middlewares.append(rlim_middleware_factory)


async def shutdown(app):
    try:
        async with app['redis_rlim'].get() as rr:
            await rr.flushdb()
    except ConnectionRefusedError:
        pass
    app['redis_rlim'].close()
    await app['redis_rlim'].wait_closed()



'''
The main web / websocket server
'''

import asyncio
from ipaddress import ip_address
import logging
from multiprocessing.managers import SyncManager
import os
import signal
import ssl

import aiohttp
from aiohttp import web
import aioredis
import aiotools
from aiopg.sa import create_engine
import uvloop
try:
    import datadog
    datadog_available = True
except ImportError:
    datadog_available = False
try:
    import raven
    raven_available = True
except ImportError:
    raven_available = False

from ai.backend.common.argparse import (
    ipaddr, path, port_no,
    host_port_pair, HostPortPair
)
from ai.backend.common.utils import env_info
from ai.backend.common.monitor import DummyDatadog, DummySentry
from ..manager import __version__
from . import GatewayStatus
from .defs import REDIS_STAT_DB
from .exceptions import (BackendError, GenericNotFound,
                         GenericBadRequest, InternalServerError)
from .admin import init as admin_init, shutdown as admin_shutdown
from .auth import init as auth_init, shutdown as auth_shutdown
from .config import load_config, init_logger
from .etcd import init as etcd_init, shutdown as etcd_shutdown
from .events import init as event_init, shutdown as event_shutdown
from .events import event_router
from .kernel import init as kernel_init, shutdown as kernel_shutdown
from .ratelimit import init as rlim_init, shutdown as rlim_shutdown

VALID_VERSIONS = frozenset([
    'v1.20160915',
    'v2.20170315',
    'v3.20170615',
])
LATEST_API_VERSION = 'v3.20170615'

log = logging.getLogger('ai.backend.gateway.server')


async def hello(request) -> web.Response:
    '''
    Returns the API version number.
    '''
    return web.json_response({'version': LATEST_API_VERSION})


async def on_prepare(request, response):
    response.headers['Server'] = 'BackendAI-API/' + LATEST_API_VERSION


async def api_middleware_factory(app, handler):
    async def api_middleware_handler(request):
        _handler = handler
        method_override = request.headers.get('X-Method-Override', None)
        if method_override:
            request = request.clone(method=method_override)
            new_match_info = await app.router.resolve(request)
            _handler = new_match_info.handler
            request._match_info = new_match_info
        if request.match_info.http_exception is not None:
            return request.match_info.http_exception
        version = int(request.match_info['version'])
        if version < 1 or version > 3:
            raise GenericBadRequest('Unsupported API major version.')
        request['api_version'] = version
        resp = (await _handler(request))
        return resp
    return api_middleware_handler


async def exception_middleware_factory(app, handler):
    async def exception_middleware_handler(request):
        try:
            app['datadog'].statsd.increment('ai.backend.gateway.api.requests')
            resp = (await handler(request))
        except BackendError as ex:
            app['sentry'].captureException()
            statsd = app['datadog'].statsd
            statsd.increment('ai.backend.gateway.api.failures')
            statsd.increment(f'ai.backend.gateway.api.status.{ex.status_code}')
            raise
        except web.HTTPException as ex:
            statsd = app['datadog'].statsd
            statsd.increment('ai.backend.gateway.api.failures')
            statsd.increment(f'ai.backend.gateway.api.status.{ex.status_code}')
            if ex.status_code == 404:
                raise GenericNotFound
            log.warning(f'Bad request: {ex!r}')
            raise GenericBadRequest
        except Exception as ex:
            app['sentry'].captureException()
            log.exception('Uncaught exception in HTTP request handlers')
            raise InternalServerError
        else:
            app['datadog'].statsd.increment(
                f'ai.backend.gateway.api.status.{resp.status}')
            return resp
    return exception_middleware_handler


async def gw_init(app):
    app.on_response_prepare.append(on_prepare)
    app.router.add_route('GET', r'/v{version:\d+}', hello)
    app['status'] = GatewayStatus.STARTING
    app['datadog'] = DummyDatadog()
    app['sentry'] = DummySentry()
    if datadog_available:
        if app.config.datadog_api_key is None:
            log.warning('datadog logging disabled (missing API key)')
        else:
            datadog.initialize(
                api_key=app.config.datadog_api_key,
                app_key=app.config.datadog_app_key)
            app['datadog'] = datadog
            log.info('datadog logging enabled')
    if raven_available:
        if app.config.raven_uri is None:
            log.info('skipping Sentry initialization due to missing DSN URI...')
        else:
            app['sentry'] = raven.Client(
                app.config.raven_uri,
                release=raven.fetch_package_version('backend.ai-manager'))
            log.info('sentry logging enabled')

    app['dbpool'] = await create_engine(
        dsn=f'host={app.config.db_addr[0]} port={app.config.db_addr[1]} '
            f'user={app.config.db_user} password={app.config.db_password} '
            f'dbname={app.config.db_name}',
        minsize=4, maxsize=16,
        echo=bool(app.config.verbose),
    )
    app['redis_stat_pool'] = await aioredis.create_pool(
        app.config.redis_addr.as_sockaddr(),
        create_connection_timeout=3.0,
        encoding='utf8',
        db=REDIS_STAT_DB)
    app.middlewares.append(exception_middleware_factory)
    app.middlewares.append(api_middleware_factory)


async def gw_shutdown(app):
    app['redis_stat_pool'].close()
    await app['redis_stat_pool'].wait_closed()
    app['dbpool'].close()
    await app['dbpool'].wait_closed()


@aiotools.actxmgr
async def server_main(loop, pidx, _args):

    app = web.Application()
    app.config = _args[0]
    app.sslctx = None
    if app.config.ssl_cert and app.config.ssl_key:
        app.sslctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        app.sslctx.load_cert_chain(str(app.config.ssl_cert),
                                   str(app.config.ssl_key))
    if app.config.service_port == 0:
        app.config.service_port = 8443 if app.sslctx else 8080

    app['shared_states'] = _args[1]
    app['pidx'] = pidx

    if pidx == 0:
        await etcd_init(app)
    await event_init(app)
    await gw_init(app)
    await auth_init(app)
    await admin_init(app)
    await rlim_init(app)
    await kernel_init(app)

    web_handler = app.make_handler()
    server = await loop.create_server(
        web_handler,
        host=str(app.config.service_ip),
        port=app.config.service_port,
        reuse_port=True,
        ssl=app.sslctx,
    )
    log.info('started.')

    try:
        yield
    finally:

        log.info('shutting down...')
        server.close()
        await server.wait_closed()

        await kernel_shutdown(app)
        await rlim_shutdown(app)
        await admin_shutdown(app)
        await auth_shutdown(app)
        await gw_shutdown(app)
        await event_shutdown(app)
        if pidx == 0:
            await etcd_shutdown(app)

        await app.shutdown()
        await web_handler.finish_connections(60.0)
        await app.cleanup()


def gw_args(parser):
    parser.add('--namespace', env_var='BACKEND_NAMESPACE',
               type=str, default='local',
               help='The namespace of this Backend.AI cluster. (default: local)')
    parser.add('--etcd-addr', env_var='BACKEND_ETCD_ADDR',
               type=host_port_pair,
               default=HostPortPair(ip_address('127.0.0.1'), 2379),
               help='The host:port pair of the etcd cluster or its proxy.')
    parser.add('--events-port', env_var='BACKEND_EVENTS_PORT',
               type=port_no, default=5002,
               help='The TCP port number where the event server listens on.')
    parser.add('--docker-registry', env_var='BACKEND_DOCKER_REGISTRY',
               type=host_port_pair, default=None,
               help='The host:port pair of the private docker registry '
                    'that caches the kernel images')
    parser.add('--heartbeat-timeout', env_var='BACKEND_HEARTBEAT_TIMEOUT',
               type=float, default=5.0,
               help='The timeout for agent heartbeats.')
    parser.add('--service-ip', env_var='BACKEND_SERVICE_IP',
               type=ipaddr, default=ip_address('0.0.0.0'),
               help='The IP where the API gateway server listens on. '
                    '(default: 0.0.0.0)')
    parser.add('--service-port', env_var='BACKEND_SERVICE_PORT',
               type=port_no, default=0,
               help='The TCP port number where the API gateway server listens on. '
                    '(default: 8080, 8443 when SSL is enabled) '
                    'To run in production, you need the root privilege '
                    'to use the standard 80/443 ports.')
    parser.add('--ssl-cert', env_var='BACKEND_SSL_CERT',
               type=path, default=None,
               help='The path to an SSL certificate file. '
                    'It may contain inter/root CA certificates as well. '
                    '(default: None)')
    parser.add('--ssl-key', env_var='BACKEND_SSL_KEY',
               type=path, default=None,
               help='The path to the private key used to make requests '
                    'for the SSL certificate. '
                    '(default: None)')
    if datadog_available:
        parser.add('--datadog-api-key', env_var='DATADOG_API_KEY',
                   type=str, default=None,
                   help='The API key for Datadog monitoring agent.')
        parser.add('--datadog-app-key', env_var='DATADOG_APP_KEY',
                   type=str, default=None,
                   help='The application key for Datadog monitoring agent.')
    if raven_available:
        parser.add('--raven-uri', env_var='RAVEN_URI',
                   type=str, default=None,
                   help='The sentry.io event report URL with DSN.')


def main():

    config = load_config(extra_args_func=gw_args)
    init_logger(config)

    log.info(f'Backend.AI Gateway {__version__}')
    log.info(f'runtime: {env_info()}')

    log_config = logging.getLogger('ai.backend.gateway.config')
    log_config.debug('debug mode enabled.')

    if config.debug:
        aiohttp.log.server_logger.setLevel('DEBUG')
        aiohttp.log.access_logger.setLevel('DEBUG')
    else:
        aiohttp.log.server_logger.setLevel('WARNING')
        aiohttp.log.access_logger.setLevel('WARNING')

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    num_workers = os.cpu_count()
    manager = SyncManager()
    manager.start(lambda: signal.signal(signal.SIGINT, signal.SIG_IGN))
    shared_states = manager.Namespace()
    shared_states.lock = manager.Lock()
    shared_states.barrier = manager.Barrier(num_workers)
    shared_states.agent_last_seen = manager.dict()

    try:
        aiotools.start_server(server_main, num_workers=num_workers,
                              extra_procs=[event_router],
                              args=(config, shared_states))
    finally:
        manager.shutdown()
        log.info('terminated.')


if __name__ == '__main__':
    main()



from ipaddress import ip_address
import logging, logging.config

from ai.backend.common.argparse import HostPortPair, host_port_pair, ipaddr, port_no
import configargparse


def load_config(argv=None, extra_args_func=None):
    parser = configargparse.ArgumentParser()
    parser.add('--agent-port', env_var='BACKEND_AGENT_PORT',
               type=port_no, default=6001,
               help='The TCP port number where the agent instances are listening on.'
                    ' (default: 6001)')
    parser.add('--redis-addr', env_var='BACKEND_REDIS_ADDR', type=host_port_pair,
               default=HostPortPair(ip_address('127.0.0.1'), 6379),
               help='The hostname-port pair of a redis server. '
                    '(default: localhost:6379)')
    parser.add('--db-addr', env_var='BACKEND_DB_ADDR', type=host_port_pair,
               default=HostPortPair(ip_address('127.0.0.1'), 5432),
               help='The hostname-port pair of a database server. '
                    '(default: localhost:5432)')
    parser.add('--db-name', env_var='BACKEND_DB_NAME',
               type=str, default='sorna',
               help='The database name. (default: sorna)')
    parser.add('--db-user', env_var='BACKEND_DB_USER',
               type=str, default='postgres',
               help='The username to authenticate to the database server. '
                    '(default: postgres)')
    parser.add('--db-password', env_var='BACKEND_DB_PASSWORD',
               type=str, default='develove',
               help='The password to authenticate to the database server. '
                    '(default: develove)')
    parser.add('--debug', env_var='BACKEND_DEBUG',
               action='store_true', default=False,
               help='Set the debug mode and verbose logging. (default: false)')
    parser.add('-v', '--verbose', env_var='BACKEND_VERBOSE',
               action='store_true', default=False,
               help='Set even more verbose logging which includes all SQL '
                    'statements issued. (default: false)')
    parser.add('--kernel-ip-override', env_var='BACKEND_KERNEL_IP_OVERRIDE',
               type=ipaddr, default=None,
               help='The IP address that overrides the actual IP address of kernel '
                    'containers when responding to our clients. '
                    'This option is used only in local development setups where all '
                    'the Backend.AI deamons and a dockerized front-end service '
                    'that uses the legacy ZMQ interface run on the same host. '
                    '"192.168.65.1" is a special IP that Docker containers '
                    '(e.g., a front-end container) use to access the host-side '
                    'services or ports opened by anonymous kernel containers.')
    if extra_args_func:
        extra_args_func(parser)
    args = parser.parse_args(args=argv)
    return args


def init_logger(config):
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'colored': {
                '()': 'coloredlogs.ColoredFormatter',
                'format': '%(asctime)s %(levelname)s %(name)s %(message)s',
                'field_styles': {'levelname': {'color': 'black', 'bold': True},
                                 'name': {'color': 'black', 'bold': True},
                                 'asctime': {'color': 'black'}},
                'level_styles': {'info': {'color': 'cyan'},
                                 'debug': {'color': 'green'},
                                 'warning': {'color': 'yellow'},
                                 'error': {'color': 'red'},
                                 'critical': {'color': 'red', 'bold': True}},
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'DEBUG',
                'formatter': 'colored',
                'stream': 'ext://sys.stderr',
            },
            'null': {
                'class': 'logging.NullHandler',
            },
        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'aiopg': {
                'handlers': ['console'],
                'propagate': False,
                'level': 'DEBUG' if config.debug else 'INFO',
            },
            'ai.backend': {
                'handlers': ['console'],
                'propagate': False,
                'level': 'DEBUG' if config.debug else 'INFO',
            },
        },
    })



import functools
import io
import re
import traceback

from aiohttp import web

_rx_sitepkg_path = re.compile(r'^.+/site-packages/')


def method_placeholder(orig_method):
    async def _handler(request):
        raise web.HTTPMethodNotAllowed(request.method, [orig_method])
    return _handler


def prettify_traceback(exc):
    # Make a compact stack trace string
    with io.StringIO() as buf:
        while exc is not None:
            print(f'Exception: {exc!r}', file=buf)
            if exc.__traceback__ is None:
                print('  (no traceback available)', file=buf)
            else:
                for frame in traceback.extract_tb(exc.__traceback__):
                    short_path = _rx_sitepkg_path.sub('<sitepkg>/', frame.filename)
                    print(f'  {short_path}:{frame.lineno} ({frame.name})', file=buf)
            exc = exc.__context__
        return f'Traceback:\n{buf.getvalue()}'


def catch_unexpected(log, raven=None):

    def _wrap(func):

        @functools.wraps(func)
        async def _wrapped(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except:
                if raven:
                    raven.captureException()
                log.exception('unexpected error!')
                raise

        return _wrapped

    return _wrap



import base64
from datetime import datetime, timedelta
import functools
import hashlib, hmac
import logging
import secrets

from aiohttp import web
from dateutil.tz import tzutc
from dateutil.parser import parse as dtparse
import simplejson as json

from .exceptions import InvalidAuthParameters, AuthorizationFailed
from .config import load_config, init_logger
from ..manager.models import keypairs

log = logging.getLogger('ai.backend.gateway.auth')


def _extract_auth_params(request):
    """
    HTTP Authorization header must be formatted as:
    "Authorization: BackendAI signMethod=HMAC-SHA256,
                    credential=<ACCESS_KEY>:<SIGNATURE>"
    """
    auth_hdr = request.headers.get('Authorization')
    if not auth_hdr:
        return None
    pieces = auth_hdr.split(' ', 1)
    if len(pieces) != 2:
        raise InvalidAuthParameters('Malformed authorization header')
    auth_type, auth_str = pieces
    if auth_type not in ('BackendAI', 'Sorna'):
        raise InvalidAuthParameters('Invalid authorization type name')

    raw_params = map(lambda s: s.strip(), auth_str.split(','))
    params = {}
    for param in raw_params:
        key, value = param.split('=', 1)
        params[key.strip()] = value.strip()

    try:
        access_key, signature = params['credential'].split(':', 1)
        ret = params['signMethod'], access_key, signature
        return ret
    except (KeyError, ValueError):
        raise InvalidAuthParameters('Missing or malformed authorization parameters')


def check_date(request) -> bool:
    raw_date = request.headers.get('Date')
    if not raw_date:
        raw_date = request.headers.get('X-BackendAI-Date',
                                       request.headers.get('X-Sorna-Date'))
    if not raw_date:
        return False
    try:
        date = dtparse(raw_date)
        if date.tzinfo is None:
            date = date.replace(tzinfo=tzutc())  # assume as UTC
        now = datetime.now(tzutc())
        min_date = now - timedelta(minutes=15)
        max_date = now + timedelta(minutes=15)
        request['date'] = date
        request['raw_date'] = raw_date
        if date < min_date or date > max_date:
            return False
    except ValueError:
        return False
    return True


async def sign_request(sign_method, request, secret_key) -> str:
    try:
        mac_type, hash_type = map(lambda s: s.lower(), sign_method.split('-'))
        assert mac_type == 'hmac', 'Unsupported request signing method (MAC type)'
        assert hash_type in hashlib.algorithms_guaranteed, \
               'Unsupported request signing method (hash type)'

        if request.has_body and not request.content_type == 'multipart/form-data':
            body = await request.read()
        else:
            body = b''
        body_hash = hashlib.new(hash_type, body).hexdigest()
        new_api_version = request.headers.get('X-BackendAI-Version')
        legacy_api_version = request.headers.get('X-Sorna-Version')
        api_version = new_api_version or legacy_api_version
        assert api_version is not None, 'API version missing in request headers'

        sign_bytes = ('{0}\n{1}\n{2}\nhost:{3}\ncontent-type:{4}\n'
                      'x-{name}-version:{5}\n{6}').format(
            request.method, str(request.rel_url), request['raw_date'],
            request.host, request.content_type, api_version,
            body_hash,
            name='backendai' if new_api_version is not None else 'sorna'
        ).encode()
        sign_key = hmac.new(secret_key.encode(),
                            request['date'].strftime('%Y%m%d').encode(),
                            hash_type).digest()
        sign_key = hmac.new(sign_key, request.host.encode(), hash_type).digest()
        return hmac.new(sign_key, sign_bytes, hash_type).hexdigest()
    except ValueError:
        raise AuthorizationFailed('Invalid signature')
    except AssertionError as e:
        raise InvalidAuthParameters(e.args[0])


async def auth_middleware_factory(app, handler):
    '''
    Fetches user information and sets up keypair, uesr, and is_authorized attributes.
    '''
    async def auth_middleware_handler(request):
        request['is_authorized'] = False
        request['is_admin'] = False
        request['keypair'] = None
        request['user'] = None
        if not check_date(request):
            raise InvalidAuthParameters('Date/time sync error')
        params = _extract_auth_params(request)
        if params:
            sign_method, access_key, signature = params
            async with app['dbpool'].acquire() as conn:
                query = (keypairs.select()
                                 .where(keypairs.c.access_key == access_key))
                result = await conn.execute(query)
                row = await result.first()
                if row is None:
                    raise AuthorizationFailed('Access key not found')
                my_signature = \
                    await sign_request(sign_method, request, row.secret_key)
                if secrets.compare_digest(my_signature, signature):
                    query = (keypairs.update()
                                     .values(last_used=datetime.now(tzutc()),
                                             num_queries=keypairs.c.num_queries + 1)
                                     .where(keypairs.c.access_key == access_key))
                    await conn.execute(query)
                    request['is_authorized'] = True
                    if row.is_admin:
                        request['is_admin'] = True
                    request['keypair'] = {
                        'access_key': access_key,
                        'concurrency_limit': row.concurrency_limit,
                        'rate_limit': row.rate_limit,
                    }
                    request['user'] = {
                        'id': row.user_id,
                    }
        # No matter if authenticated or not, pass-through to the handler.
        # (if it's required, auth_required decorator will handle the situation.)
        return (await handler(request))
    return auth_middleware_handler


def auth_required(handler):
    @functools.wraps(handler)
    async def wrapped(request):
        if request.get('is_authorized', False):
            return (await handler(request))
        raise AuthorizationFailed
    return wrapped


def admin_required(handler):
    @functools.wraps(handler)
    async def wrapped(request):
        if request.get('is_authorized', False) and request.get('is_admin', False):
            return (await handler(request))
        raise AuthorizationFailed
    return wrapped


@auth_required
async def authorize(request) -> web.Response:
    try:
        params = json.loads(await request.text())
    except json.decoder.JSONDecodeError:
        raise InvalidAuthParameters('Malformed body')
    resp_data = {'authorized': 'yes'}
    if 'echo' in params:
        resp_data['echo'] = params['echo']
    return web.json_response(resp_data)


def generate_keypair():
    '''
    AWS-like access key and secret key generation.
    '''
    ak = 'AKIA' + base64.b32encode(secrets.token_bytes(10)).decode('ascii')
    sk = secrets.token_urlsafe(30)
    return ak, sk


async def init(app):
    app.router.add_route('GET', r'/v{version:\d+}/authorize', authorize)
    app.middlewares.append(auth_middleware_factory)


async def shutdown(app):
    pass


if __name__ == '__main__':

    def auth_args(parser):
        parser.add('--generate-keypair', action='store_true', default=False,
                   help='Generate a pair of access key and secret key.')

    config = load_config(extra_args_func=auth_args)
    init_logger(config)

    if config.generate_keypair:
        ak, sk = generate_keypair()
        print(f'Access Key: {ak} ({len(ak)} bytes)')
        print(f'Secret Key: {sk} ({len(sk)} bytes)')



'''
This module defines a series of Backend.AI-specific errors based on HTTP Error
classes from aiohttp.
Raising a BackendError automatically is automatically mapped to a corresponding
HTTP error response with RFC7807-style JSON-encoded description in its response
body.

In the client side, you should use "type" field in the body to distinguish
canonical error types beacuse "title" field may change due to localization and
future UX improvements.
'''

from aiohttp import web
import simplejson as json


class BackendError(web.HTTPError):
    '''
    An RFC-7807 error class as a drop-in replacement of the original
    aiohttp.web.HTTPError subclasses.
    '''

    status_code = 500
    error_type  = 'https://api.backend.ai/probs/general-error'
    error_title = 'General Backend API Error.'

    def __init__(self, extra_msg=None):
        super().__init__()
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.content_type = 'application/problem+json'
        if extra_msg:
            self.error_title += f' ({extra_msg})'
        self.body = json.dumps({
            'type': self.error_type,
            'title': self.error_title,
        }).encode()


class GenericNotFound(web.HTTPNotFound, BackendError):
    error_type  = 'https://api.backend.ai/probs/generic-not-found'
    error_title = 'Unknown URL path.'


class GenericBadRequest(web.HTTPBadRequest, BackendError):
    error_type  = 'https://api.backend.ai/probs/generic-bad-request'
    error_title = 'Bad request.'


class InternalServerError(web.HTTPInternalServerError, BackendError):
    error_type  = 'https://api.backend.ai/probs/internal-server-error'
    error_title = 'Internal server error.'


class ServiceUnavailable(web.HTTPServiceUnavailable, BackendError):
    error_type  = 'https://api.backend.ai/probs/service-unavailable'
    error_title = 'Serivce unavailable.'


class QueryNotImplemented(web.HTTPServiceUnavailable, BackendError):
    error_type  = 'https://api.backend.ai/probs/not-implemented'
    error_title = 'This API query is not implemented.'


class InvalidAuthParameters(web.HTTPBadRequest, BackendError):
    error_type  = 'https://api.backend.ai/probs/invalid-auth-params'
    error_title = 'Missing or invalid authorization parameters.'


class AuthorizationFailed(web.HTTPUnauthorized, BackendError):
    error_type  = 'https://api.backend.ai/probs/auth-failed'
    error_title = 'Credential/signature mismatch.'


class InvalidAPIParameters(web.HTTPBadRequest, BackendError):
    error_type  = 'https://api.backend.ai/probs/invalid-api-params'
    error_title = 'Missing or invalid API parameters.'


class InstanceNotFound(web.HTTPNotFound, BackendError):
    error_type  = 'https://api.backend.ai/probs/instance-not-found'
    error_title = 'No such instance.'


class KernelNotFound(web.HTTPNotFound, BackendError):
    error_type  = 'https://api.backend.ai/probs/kernel-not-found'
    error_title = 'No such kernel.'


class QuotaExceeded(web.HTTPPreconditionFailed, BackendError):
    error_type  = 'https://api.backend.ai/probs/quota-exceeded'
    error_title = 'You have reached your resource limit.'


class RateLimitExceeded(web.HTTPTooManyRequests, BackendError):
    error_type  = 'https://api.backend.ai/probs/rate-limit-exceeded'
    error_title = 'You have reached your API query rate limit.'


class InstanceNotAvailable(web.HTTPServiceUnavailable, BackendError):
    error_type  = 'https://api.backend.ai/probs/instance-not-available'
    error_title = 'There is no available instance.'


class AgentError(RuntimeError):
    '''
    A dummy exception class to distinguish agent-side errors passed via
    aiozmq.rpc calls.

    It carrise two args tuple: the exception type and exception arguments from
    the agent.
    '''
    pass


class BackendAgentError(BackendError):
    '''
    An RFC-7807 error class that wraps agent-side errors.
    '''

    _short_type_map = {
        'TIMEOUT': 'https://api.backend.ai/probs/agent-timeout',
        'INVALID_INPUT': 'https://api.backend.ai/probs/agent-invalid-input',
        'FAILURE': 'https://api.backend.ai/probs/agent-failure',
    }

    def __init__(self, agent_error_type, exc_info=None):
        super().__init__()
        if not agent_error_type.startswith('https://'):
            agent_error_type = self._short_type_map[agent_error_type.upper()]
        self.args = (
            self.status_code,
            self.reason,
            self.error_type,
            agent_error_type,
        )
        if isinstance(exc_info, str):
            agent_error_title = exc_info
            agent_details = {
                'type': agent_error_type,
                'title': agent_error_title,
            }
        elif isinstance(exc_info, AgentError):
            if isinstance(exc_info.args[0], Exception):
                inner_name = type(exc_info.args[0]).__name__
            elif (isinstance(exc_info.args[0], type) and
                  issubclass(exc_info.args[0], Exception)):
                inner_name = exc_info.args[0].__name__
            else:
                inner_name = str(exc_info.args[0])
            inner_args = ', '.join(repr(a) for a in exc_info.args[1])
            agent_error_title = 'Agent-side exception occurred.'
            agent_details = {
                'type': agent_error_type,
                'title': agent_error_title,
                'exception': f"{inner_name}({inner_args})",
            }
        elif isinstance(exc_info, Exception):
            agent_error_title = 'Unexpected exception ocurred.'
            agent_details = {
                'type': agent_error_type,
                'title': agent_error_title,
                'exception': repr(exc_info),
            }
        else:
            agent_error_title = None if exc_info is None else str(exc_info)
            agent_details = {
                'type': agent_error_type,
                'title': agent_error_title,
            }
        self.agent_error_type = agent_error_type
        self.agent_error_title = agent_error_title
        self.body = json.dumps({
            'type': self.error_type,
            'title': self.error_title,
            'agent-details': agent_details,
        }).encode()


class KernelCreationFailed(web.HTTPInternalServerError, BackendAgentError):
    error_type  = 'https://api.backend.ai/probs/kernel-creation-failed'
    error_title = 'Kernel creation has failed.'


class KernelDestructionFailed(web.HTTPInternalServerError, BackendAgentError):
    error_type  = 'https://api.backend.ai/probs/kernel-destruction-failed'
    error_title = 'Kernel destruction has failed.'


class KernelRestartFailed(web.HTTPInternalServerError, BackendAgentError):
    error_type  = 'https://api.backend.ai/probs/kernel-restart-failed'
    error_title = 'Kernel restart has failed.'


class KernelExecutionFailed(web.HTTPInternalServerError, BackendAgentError):
    error_type  = 'https://api.backend.ai/probs/kernel-execution-failed'
    error_title = 'Executing user code in the kernel has failed.'



from setuptools import setup
from pathlib import Path
import ai.backend.manager


requires = [
    'aioconsole>=0.1.3',
    'aiodataloader',
    'aiohttp~=2.2.0',
    'aiopg~=0.13.0',
    'aioredis~=0.3.3',
    'aiotools>=0.4.5',
    'aiozmq>=0.7',
    'alembic~=0.9.2',
    'coloredlogs>=5.2',
    'ConfigArgParse',
    'graphene>=2.0.dev',
    'iso8601',  # required by graphene
    'namedlist',
    'psycopg2~=2.7.0',
    'python-dateutil>=2.5',
    'pyzmq>=16.0',
    'simplejson',
    'SQLAlchemy~=1.1.14',
    'uvloop>=0.8',
    'backend.ai-common~=1.0.3',
]
build_requires = [
    'wheel',
    'twine',
]
test_requires = [
    'pytest>=3.1',
    'pytest-asyncio',
    'pytest-aiohttp',
    'pytest-cov',
    'pytest-mock',
    'codecov',
    'flake8',
]
dev_requires = build_requires + test_requires + [
    'pytest-sugar',
]
ci_requires = []
monitor_requires = [
    'datadog>=0.16.0',
    'raven>=6.1',
]


setup(
    name='backend.ai-manager',
    version=ai.backend.manager.__version__,
    description='Backend.AI Manager',
    long_description=Path('README.rst').read_text(),
    url='https://github.com/lablup/backend.ai-manager',
    author='Lablup Inc.',
    author_email='joongi@lablup.com',
    license='LGPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
    packages=[
        'ai.backend.manager',
        'ai.backend.gateway',
    ],
    python_requires='>=3.6',
    install_requires=requires,
    extras_require={
        'build': build_requires,
        'test': test_requires,
        'dev': dev_requires,
        'ci': ci_requires,
        'monitor': monitor_requires,
    },
    data_files=[],
)



import asyncio
from datetime import datetime
from dateutil.tz import tzutc
import logging
import sys
import uuid

import aiozmq, aiozmq.rpc
from aiozmq.rpc.base import GenericError, NotFoundError, ParametersError
import aiotools
from async_timeout import timeout as _timeout
import sqlalchemy as sa
import zmq

from ..gateway.exceptions import (
    BackendError,
    InstanceNotAvailable, InstanceNotFound, KernelNotFound,
    KernelCreationFailed, KernelDestructionFailed,
    KernelExecutionFailed, KernelRestartFailed,
    AgentError)
from .models import (
    agents, kernels, keypairs,
    ResourceSlot, AgentStatus, KernelStatus
)
from ..gateway.utils import catch_unexpected

__all__ = ['InstanceRegistry', 'InstanceNotFound']

log = logging.getLogger('ai.backend.manager.registry')


@aiotools.actxmgr
async def RPCContext(addr, timeout=10):
    preserved_exceptions = (
        NotFoundError,
        ParametersError,
        asyncio.TimeoutError,
        asyncio.CancelledError,
        asyncio.InvalidStateError,
    )
    server = None
    try:
        server = await aiozmq.rpc.connect_rpc(
            connect=addr, error_table={
                'concurrent.futures._base.TimeoutError': asyncio.TimeoutError,
            })
        server.transport.setsockopt(zmq.LINGER, 50)
        with _timeout(timeout):
            yield server
    except:
        exc_type, exc, tb = sys.exc_info()
        if issubclass(exc_type, GenericError):
            e = AgentError(exc.args[0], exc.args[1])
            raise e.with_traceback(tb)
        elif issubclass(exc_type, preserved_exceptions):
            raise
        else:
            e = AgentError(exc_type, exc.args)
            raise e.with_traceback(tb)
    finally:
        if server:
            server.close()


@aiotools.actxmgr
async def reenter_txn(pool, conn):
    if conn is None:
        async with pool.acquire() as conn, conn.begin():
            yield conn
    else:
        yield conn


# TODO: rename to AgentRegistry
class InstanceRegistry:
    '''
    Provide a high-level API to create, destroy, and query the computation
    kernels.

    The registry is also responsible to implement our resource management
    policy, such as the limitation of maximum number of kernels per instance.
    '''

    def __init__(self, dbpool, redis_stat_pool, loop=None):
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.dbpool = dbpool
        self.redis_stat_pool = redis_stat_pool

    async def init(self):
        pass

    async def terminate(self):
        pass

    async def get_instance(self, inst_id, field=None):
        async with self.dbpool.acquire() as conn:
            query = (sa.select(['id', field] if field else None)
                       .select_from(agents)
                       .where(agents.c.id == inst_id))
            result = await conn.execute(query)
            row = await result.first()
            if not row:
                raise InstanceNotFound(inst_id)
            return row

    async def enumerate_instances(self, check_shadow=True):
        async with self.dbpool.acquire() as conn:
            query = (sa.select('*').select_from(agents))
            if check_shadow:
                query = query.where(agents.c.status == AgentStatus.ALIVE)
            async for row in conn.execute(query):
                yield row

    async def update_instance(self, inst_id, updated_fields):
        async with self.dbpool.acquire() as conn:
            query = (sa.update(agents)
                       .values(**updated_fields)
                       .where(agents.c.id == inst_id))
            await conn.execute(query)

    @aiotools.actxmgr
    async def handle_kernel_exception(
            self, op, sess_id,
            error_callback=None,
            cancellation_callback=None,
            set_error=False):
        op_exc = {
            'create_kernel': KernelCreationFailed,
            'restart_kernel': KernelRestartFailed,
            'destroy_kernel': KernelDestructionFailed,
            'execute': KernelExecutionFailed,
            'upload_file': KernelExecutionFailed,
        }
        exc_class = op_exc[op]
        try:
            yield
        except asyncio.TimeoutError:
            if set_error:
                await self.set_kernel_status(sess_id, KernelStatus.ERROR,
                                             status_info=f'Operation timeout ({op})')
            if error_callback:
                await error_callback()
            raise exc_class('TIMEOUT')
        except asyncio.CancelledError:
            if cancellation_callback:
                await cancellation_callback()
            raise
        except AgentError as e:
            log.exception(f'{op}: agent-side error')
            if set_error:
                await self.set_kernel_status(sess_id, KernelStatus.ERROR,
                                             status_info='Agent error')
            if error_callback:
                await error_callback()
            raise exc_class('FAILURE', e)
        except BackendError:
            # silently re-raise to make them handled by gateway http handlers
            raise
        except:
            log.exception(f'{op}: other error')
            # TODO: raven.captureException()
            if set_error:
                await self.set_kernel_status(sess_id, KernelStatus.ERROR,
                                             status_info='Other error')
            if error_callback:
                await error_callback()
            raise

    async def get_kernel(self, kern_id: uuid.UUID, field=None, allow_stale=False):
        '''
        Retreive the kernel information from the given kernel ID.
        This ID is unique for all individual agent-spawned containers.

        If ``field`` is given, it extracts only the raw value of the given
        field, without wrapping it as Kernel object.
        If ``allow_stale`` is true, it skips checking validity of the kernel
        owner instance.
        '''

        cols = [kernels.c.id, kernels.c.sess_id,
                kernels.c.agent_addr, kernels.c.access_key]
        if field == '*':
            cols = '*'
        elif isinstance(field, (tuple, list)):
            cols.extend(field)
        elif isinstance(field, (sa.Column, sa.sql.elements.ColumnClause)):
            cols.append(field)
        elif isinstance(field, str):
            cols.append(sa.column(field))
        async with self.dbpool.acquire() as conn:
            if allow_stale:
                query = (sa.select(cols)
                           .select_from(kernels)
                           .where(kernels.c.id == kern_id)
                           .limit(1).offset(0))
            else:
                query = (sa.select(cols)
                           .select_from(kernels.join(agents))
                           .where((kernels.c.id == kern_id) &
                                  (kernels.c.status.in_([KernelStatus.BUILDING,
                                                         KernelStatus.RUNNING])) &
                                  (agents.c.status == AgentStatus.ALIVE) &
                                  (agents.c.id == kernels.c.agent))
                           .limit(1).offset(0))
            result = await conn.execute(query)
            row = await result.first()
            if row is None:
                raise KernelNotFound
            return row

    async def get_kernel_session(self, sess_id: str, field=None, allow_stale=False):
        '''
        Retreive the kernel information from the session ID (client-side
        session token).  If the kernel is composed of multiple containers, it
        returns the address of the master container.

        If ``field`` is given, it extracts only the raw value of the given
        field, without wrapping it as Kernel object.  If ``allow_stale`` is
        true, it skips checking validity of the kernel owner instance.
        '''

        cols = [kernels.c.id, kernels.c.sess_id, kernels.c.lang,
                kernels.c.agent_addr, kernels.c.access_key]
        if field == '*':
            cols = '*'
        elif isinstance(field, (tuple, list)):
            cols.extend(field)
        elif isinstance(field, (sa.Column, sa.sql.elements.ColumnClause)):
            cols.append(field)
        elif isinstance(field, str):
            cols.append(sa.column(field))
        async with self.dbpool.acquire() as conn:
            if allow_stale:
                query = (sa.select(cols)
                           .select_from(kernels)
                           .where((kernels.c.sess_id == sess_id) &
                                  (kernels.c.role == 'master'))
                           .limit(1).offset(0))
            else:
                query = (sa.select(cols)
                           .select_from(kernels.join(agents))
                           .where((kernels.c.sess_id == sess_id) &
                                  (kernels.c.role == 'master') &
                                  (kernels.c.status.in_([KernelStatus.BUILDING,
                                                         KernelStatus.RUNNING])) &
                                  (agents.c.status == AgentStatus.ALIVE) &
                                  (agents.c.id == kernels.c.agent))
                           .limit(1).offset(0))
            result = await conn.execute(query)
            row = await result.first()
            if row is None:
                raise KernelNotFound
            return row

    async def get_kernels(self, kern_ids, field=None, allow_stale=False):
        '''
        Batched version of :meth:`get_kernel() <InstanceRegistry.get_kernel>`.
        The order of the returend array is same to the order of ``kern_ids``.
        For non-existent or missing kernel IDs, it fills None in their
        positions without raising KernelNotFound exception.
        '''

        cols = [kernels.c.id, kernels.c.sess_id,
                kernels.c.agent_addr, kernels.c.access_key]
        if isinstance(field, (tuple, list)):
            cols.extend(field)
        elif isinstance(field, (sa.Column, sa.sql.elements.ColumnClause)):
            cols.append(field)
        elif isinstance(field, str):
            cols.append(sa.column(field))
        async with self.dbpool.acquire() as conn:
            if allow_stale:
                query = (sa.select(cols)
                           .select_from(kernels)
                           .where((kernels.c.sess_id.in_(kern_ids)) &
                                  (kernels.c.role == 'master')))
            else:
                query = (sa.select(cols)
                           .select_from(kernels.join(agents))
                           .where((kernels.c.sess_id.in_(kern_ids)) &
                                  (kernels.c.role == 'master') &
                                  (agents.c.status == AgentStatus.ALIVE) &
                                  (agents.c.id == kernels.c.agent)))
            result = await conn.execute(query)
            rows = await result.fetchall()
            return rows

    async def set_kernel_status(self, sess_id, status,
                                db_connection=None, **extra_fields):
        data = {
            'status': status,
        }
        data.update(extra_fields)
        async with reenter_txn(self.dbpool, db_connection) as conn:
            query = (sa.update(kernels)
                       .values(data)
                       .where(kernels.c.sess_id == sess_id))
            await conn.execute(query)

    @catch_unexpected(log)
    async def get_or_create_kernel(self, client_sess_token, lang, owner_access_key,
                                   limits=None, mounts=None, conn=None):
        assert owner_access_key
        try:
            kern = await self.get_kernel_session(client_sess_token)
            created = False
        except KernelNotFound:
            kern = await self.create_kernel(
                client_sess_token, lang, owner_access_key,
                limits=limits, mounts=mounts, conn=conn)
            created = True
        assert kern is not None
        return kern, created

    def get_kernel_slot(self, lang):
        # TODO: implement
        return ResourceSlot(
            None, 1, 1, 0
        )

    async def create_kernel(self, sess_id, lang, owner_access_key,
                            limits=None, mounts=None, conn=None):
        agent_id = None
        limits = limits or {}
        mounts = mounts or []
        required_slot = self.get_kernel_slot(lang)
        created_info = None

        async with reenter_txn(self.dbpool, conn) as conn:

            # scan available slots from alive agents
            avail_slots = []
            query = (sa.select([agents], for_update=True)
                       .where(agents.c.status == AgentStatus.ALIVE))
            async for row in conn.execute(query):
                sdiff = ResourceSlot(
                    row['id'],
                    row['mem_slots'] - row['used_mem_slots'],
                    row['cpu_slots'] - row['used_cpu_slots'],
                    row['gpu_slots'] - row['used_gpu_slots'],
                )
                avail_slots.append(sdiff)

            # check minimum requirement
            avail_slots = [s for s in avail_slots
                           if s.mem >= required_slot.mem and
                              s.cpu >= required_slot.cpu and
                              s.gpu >= required_slot.gpu]

            # load-balance
            if avail_slots:
                agent_id = (max(avail_slots, key=lambda s: s.mem + s.cpu + s.gpu)).id
            else:
                raise InstanceNotAvailable

            # reserve slots
            mem_col = agents.c.used_mem_slots
            cpu_col = agents.c.used_cpu_slots
            gpu_col = agents.c.used_gpu_slots
            query = (sa.update(agents)
                       .values({
                           'used_mem_slots': mem_col + required_slot.mem,
                           'used_cpu_slots': cpu_col + required_slot.cpu,
                           'used_gpu_slots': gpu_col + required_slot.gpu,
                       })
                       .where(agents.c.id == agent_id))
            result = await conn.execute(query)
            assert result.rowcount == 1

            # Create kernel by invoking the agent on the instance.
            query = (sa.select([agents.c.addr])
                       .where(agents.c.id == agent_id))
            agent_addr = await conn.scalar(query)
            assert agent_addr is not None

            # Prepare kernel.
            kernel_id = uuid.uuid4()
            query = kernels.insert().values({
                'id': kernel_id,
                'status': KernelStatus.PREPARING,
                'sess_id': sess_id,
                'role': 'master',
                'agent': agent_id,
                'agent_addr': agent_addr,
                'access_key': owner_access_key,
                'lang': lang,
                'mem_slot': required_slot.mem,
                'cpu_slot': required_slot.cpu,
                'gpu_slot': required_slot.gpu,
                'cpu_set': [],
                'gpu_set': [],
                'repl_in_port': 0,
                'repl_out_port': 0,
                'stdin_port': 0,
                'stdout_port': 0,
            })
            result = await conn.execute(query)
            assert result.rowcount == 1

            async with self.handle_kernel_exception('create_kernel', sess_id):
                async with RPCContext(agent_addr, 3) as rpc:
                    config = {
                        'lang': lang,
                        # TODO: apply limits/mounts/vfolders
                    }
                    created_info = await rpc.call.create_kernel(str(kernel_id),
                                                                config)
                if created_info is None:
                    raise KernelCreationFailed('ooops')
                log.debug(f'create_kernel("{sess_id}") -> '
                          f'created on {agent_id}\n{created_info!r}')
                assert str(kernel_id) == created_info['id']
                kernel_access_info = {
                    'id': kernel_id,
                    'sess_id': sess_id,
                    'agent': agent_id,
                    'agent_addr': agent_addr,
                }
                query = (kernels.update()
                                .values({
                                    'status': KernelStatus.RUNNING,
                                    'container_id': created_info['container_id'],
                                    'cpu_set': list(created_info['cpu_set']),
                                    'gpu_set': list(created_info['gpu_set']),
                                    'repl_in_port': created_info['repl_in_port'],
                                    'repl_out_port': created_info['repl_out_port'],
                                    'stdin_port': created_info['stdin_port'],
                                    'stdout_port': created_info['stdout_port'],
                                })
                                .where(kernels.c.id == kernel_id))
                result = await conn.execute(query)
                assert result.rowcount == 1
                return kernel_access_info

    async def destroy_kernel(self, sess_id):
        log.debug(f"destroy_kernel({sess_id})")
        async with self.handle_kernel_exception('destroy_kernel', sess_id,
                                                set_error=True):
            kernel = await self.get_kernel_session(sess_id)
            await self.set_kernel_status(sess_id, KernelStatus.TERMINATING)
            async with RPCContext(kernel['agent_addr'], 10) as rpc:
                await rpc.call.destroy_kernel(str(kernel['id']))

    async def restart_kernel(self, sess_id):
        log.debug(f'restart_kernel({sess_id})')
        async with self.handle_kernel_exception('restart_kernel', sess_id,
                                                set_error=True):
            extra_cols = (kernels.c.lang, kernels.c.cpu_set, kernels.c.gpu_set)
            kernel = await self.get_kernel_session(sess_id, extra_cols)
            await self.set_kernel_status(sess_id, KernelStatus.RESTARTING)
            async with RPCContext(kernel['agent_addr'], 30) as rpc:
                prev_config = {
                    'lang': kernel['lang'],
                    'cpu_set': kernel['cpu_set'],
                    'gpu_set': kernel['gpu_set'],
                }
                kernel_info = await rpc.call.restart_kernel(str(kernel['id']),
                                                            prev_config)
            # TODO: what if prev status was "building" or others?
            await self.set_kernel_status(sess_id, KernelStatus.RUNNING,
                                         container_id=kernel_info['container_id'],
                                         cpu_set=list(kernel_info['cpu_set']),
                                         gpu_set=list(kernel_info['gpu_set']),
                                         repl_in_port=kernel_info['repl_in_port'],
                                         repl_out_port=kernel_info['repl_out_port'],
                                         stdin_port=kernel_info['stdin_port'],
                                         stdout_port=kernel_info['stdout_port'])

    async def execute(self, sess_id, api_version, run_id, mode, code, opts):
        log.debug(f'execute:v{api_version}({sess_id}, {mode})')
        async with self.handle_kernel_exception('execute', sess_id):
            kernel = await self.get_kernel_session(sess_id)
            # The agent aggregates at most 2 seconds of outputs
            # if the kernel runs for a long time.
            async with RPCContext(kernel['agent_addr'], 300) as rpc:
                exec_coro = rpc.call.execute(api_version, str(kernel['id']),
                                             run_id, mode, code, opts)
                if exec_coro is None:
                    raise RuntimeError('execute cancelled')
                try:
                    return await exec_coro
                except TypeError as e:
                    log.exception('typeerror????')

    async def interrupt_kernel(self, sess_id):
        log.debug(f'interrupt({sess_id})')
        async with self.handle_kernel_exception('execute', sess_id):
            kernel = await self.get_kernel_session(sess_id)
            async with RPCContext(kernel['agent_addr'], 5) as rpc:
                exec_coro = rpc.call.interrupt_kernel(str(kernel['id']))
                if exec_coro is None:
                    raise RuntimeError('interrupt cancelled')
                return await exec_coro

    async def get_completions(self, sess_id, mode, text, opts):
        log.debug(f'get_completions({sess_id}, {mode})')
        async with self.handle_kernel_exception('execute', sess_id):
            kernel = await self.get_kernel_session(sess_id)
            async with RPCContext(kernel['agent_addr'], 5) as rpc:
                exec_coro = rpc.call.get_completions(str(kernel['id']), mode,
                                                     text, opts)
                if exec_coro is None:
                    raise RuntimeError('get_completions cancelled')
                return await exec_coro

    async def upload_file(self, sess_id, filename, payload):
        log.debug(f'upload_file({sess_id}, {filename})')
        async with self.handle_kernel_exception('upload_file', sess_id):
            kernel = await self.get_kernel_session(sess_id)
            async with RPCContext(kernel['agent_addr'], 180) as rpc:
                result = \
                    await rpc.call.upload_file(str(kernel['id']), filename, payload)
                return result

    async def update_kernel(self, sess_id, updated_fields, conn=None):
        log.debug(f'update_kernel({sess_id})')
        async with reenter_txn(self.dbpool, conn) as conn:
            query = (sa.update(kernels)
                       .values(updated_fields)
                       .where((kernels.c.sess_id == sess_id) &
                              (kernels.c.role == 'master')))
            await conn.execute(query)

    async def increment_session_usage(self, sess_id, conn=None):
        log.debug(f'increment_session_usage({sess_id})')
        async with reenter_txn(self.dbpool, conn) as conn:
            query = (sa.update(kernels)
                       .values(num_queries=kernels.c.num_queries + 1)
                       .where((kernels.c.sess_id == sess_id) &
                              (kernels.c.role == 'master')))
            await conn.execute(query)

    async def get_kernels_in_instance(self, inst_id, conn=None):
        async with reenter_txn(self.dbpool, conn) as conn:
            query = (sa.select([kernels.c.sess_id])
                       .select_from(kernels)
                       .where(kernels.c.agent == inst_id))
            result = await conn.execute(query)
            rows = await result.fetchall()
            if not rows:
                return tuple()
            return rows

    async def handle_heartbeat(self, agent_id, agent_info):
        async with self.dbpool.acquire() as conn:
            # TODO: check why sa.column('status') does not work
            query = (sa.select([agents.c.status], for_update=True)
                       .select_from(agents)
                       .where(agents.c.id == agent_id))
            prev_status = await conn.scalar(query)
            if prev_status is None:
                # new agent detected!
                log.info(f'agent {agent_id} joined!')
                query = agents.insert().values({
                    'id': agent_id,
                    'status': AgentStatus.ALIVE,
                    'region': agent_info['region'],
                    'mem_slots': agent_info['mem_slots'],
                    'cpu_slots': agent_info['cpu_slots'],
                    'gpu_slots': agent_info['gpu_slots'],
                    'used_mem_slots': 0,
                    'used_cpu_slots': 0,
                    'used_gpu_slots': 0,
                    'addr': agent_info['addr'],
                    'first_contact': datetime.now(tzutc()),
                    'lost_at': None,
                })
                result = await conn.execute(query)
                assert result.rowcount == 1
            elif prev_status == AgentStatus.ALIVE:
                pass
            elif prev_status in (AgentStatus.LOST, AgentStatus.TERMINATED):
                log.warning(f'agent {agent_id} revived!')
                query = (sa.update(agents)
                           .values({
                               'status': AgentStatus.ALIVE,
                               'lost_at': None,
                           })
                           .where(agents.c.id == agent_id))
                await conn.execute(query)
            else:
                log.error(f'should not reach here! {type(prev_status)}')

    async def mark_agent_terminated(self, agent_id, status, conn=None):
        # TODO: interpret kern_id to sess_id
        # for kern_id in (await app['registry'].get_kernels_in_instance(agent_id)):
        #     for handler in app['stream_pty_handlers'][kern_id].copy():
        #         handler.cancel()
        #         await handler
        #  TODO: define behavior when agent reuse running instances upon revive
        # await app['registry'].forget_all_kernels_in_instance(agent_id)
        async with reenter_txn(self.dbpool, conn) as conn:

            query = (sa.select([agents.c.status])
                       .select_from(agents)
                       .where(agents.c.id == agent_id))
            result = await conn.execute(query)
            prev_status = await result.scalar()
            if prev_status in (None, AgentStatus.LOST, AgentStatus.TERMINATED):
                return

            if status == AgentStatus.LOST:
                log.warning(f'agent {agent_id} heartbeat timeout detected.')
            elif status == AgentStatus.TERMINATED:
                log.info(f'agent {agent_id} has terminated.')
            query = (sa.update(agents)
                       .values({
                           'status': status,
                           'lost_at': datetime.now(tzutc()),
                       })
                       .where(agents.c.id == agent_id))
            await conn.execute(query)

    async def mark_kernel_terminated(self, kernel_id, conn=None):
        '''
        Mark the kernel (individual worker) terminated and release
        the resource slots occupied by it.
        '''
        async with reenter_txn(self.dbpool, conn) as conn:
            # check if already terminated
            query = (sa.select([kernels.c.status])
                       .select_from(kernels)
                       .where(kernels.c.id == kernel_id))
            result = await conn.execute(query)
            prev_status = await result.scalar()
            if prev_status in (None, KernelStatus.TERMINATED):
                return

            # change the status to TERMINATED
            # (we don't delete the row for later logging and billing)
            kern_data = {
                'status': KernelStatus.TERMINATED,
                'terminated_at': datetime.now(tzutc()),
            }
            async with self.redis_stat_pool.get() as rs:
                kern_stat = await rs.hgetall(kernel_id)
                if kern_stat is not None:
                    kern_data.update({
                        'cpu_used': int(float(kern_stat['cpu_used'])),
                        'mem_max_bytes': int(kern_stat['mem_max_bytes']),
                        'net_rx_bytes': int(kern_stat['net_rx_bytes']),
                        'net_tx_bytes': int(kern_stat['net_tx_bytes']),
                        'io_read_bytes': int(kern_stat['io_read_bytes']),
                        'io_write_bytes': int(kern_stat['io_write_bytes']),
                        'io_max_scratch_size': int(kern_stat['io_max_scratch_size']),
                    })
            query = (sa.update(kernels)
                       .values(kern_data)
                       .where(kernels.c.id == kernel_id))
            await conn.execute(query)

            # release resource slots
            query = (sa.select([sa.column('agent'),
                                sa.column('mem_slot'),
                                sa.column('cpu_slot'),
                                sa.column('gpu_slot')])
                       .select_from(kernels)
                       .where(kernels.c.id == kernel_id))
            result = await conn.execute(query)
            kernel = await result.first()
            if kernel is None:
                return
            mem_col = agents.c.used_mem_slots
            cpu_col = agents.c.used_cpu_slots
            gpu_col = agents.c.used_gpu_slots
            query = (sa.update(agents)
                       .values({
                           'used_mem_slots': mem_col - kernel['mem_slot'],
                           'used_cpu_slots': cpu_col - kernel['cpu_slot'],
                           'used_gpu_slots': gpu_col - kernel['gpu_slot'],
                       })
                       .where(agents.c.id == kernel['agent']))
            await conn.execute(query)

    async def mark_session_terminated(self, sess_id):
        '''
        Mark the compute session terminated and restore the concurrency limit
        of the owner access key.  Releasing resource limits is handled by
        func:`mark_kernel_terminated`.
        '''
        async with self.dbpool.acquire() as conn:

            # restore concurrency usage of the owner access-key
            query = (sa.select([sa.column('id'),
                                sa.column('access_key')])
                       .select_from(kernels)
                       .where(kernels.c.sess_id == sess_id))
            result = await conn.execute(query)
            all_kernels = await result.fetchall()
            num_kernels = len(all_kernels)
            if num_kernels > 0:
                access_key = all_kernels[0]['access_key']
                # concurrency is per session.
                query = (sa.update(keypairs)
                           .values({
                               'concurrency_used': (keypairs.c.concurrency_used - 1),
                           })
                           .where(keypairs.c.access_key == access_key))
                await conn.execute(query)

    async def forget_instance(self, inst_id):
        async with self.dbpool.acquire() as conn:
            query = (sa.update(agents)
                       .values(status=AgentStatus.TERMINATED,
                               lost_at=datetime.now(tzutc()))
                       .where(agents.c.id == inst_id))
            await conn.execute(query)



import asyncio
from collections import defaultdict
import functools
import logging
import sys

import aiozmq
import zmq

from ai.backend.common import msgpack

log = logging.getLogger('ai.backend.gateway.events')

if sys.platform == 'darwin':
    EVENT_IPC_ADDR = 'ipc://ai.backend.agent-events'
elif sys.platform == 'linux':
    EVENT_IPC_ADDR = 'ipc://@/ai.backend.agent-events'
else:
    raise NotImplementedError('unsupported platform')


def event_router(_, pidx, args):
    # run as extra_procs by aiotools
    ctx = zmq.Context()
    ctx.linger = 50
    in_sock = ctx.socket(zmq.PULL)
    in_sock.bind(f'tcp://*:{args[0].events_port}')
    out_sock = ctx.socket(zmq.PUSH)
    out_sock.bind(EVENT_IPC_ADDR)
    try:
        zmq.proxy(in_sock, out_sock)
    except KeyboardInterrupt:
        pass
    except:
        log.exception('unexpected error')
        # raven.captureException()
    finally:
        in_sock.close()
        out_sock.close()
        ctx.term()


class EventDispatcher:

    def __init__(self, app, loop=None):
        self.app = app
        self.loop = loop if loop else asyncio.get_event_loop()
        self.handlers = defaultdict(list)

    def add_handler(self, event_name, callback):
        assert callable(callback)
        self.handlers[event_name].append(callback)

    def dispatch(self, event_name, agent_id, args=tuple()):
        first_arg = f', {args[0]}' if args else ''
        log.debug(f"DISPATCH({event_name}/{agent_id}{first_arg})")
        for handler in self.handlers[event_name]:
            if asyncio.iscoroutine(handler) or asyncio.iscoroutinefunction(handler):
                self.loop.create_task(handler(self.app, agent_id, *args))
            else:
                cb = functools.partial(handler, self.app, agent_id, *args)
                self.loop.call_soon(cb)


async def event_subscriber(dispatcher):
    event_sock = await aiozmq.create_zmq_stream(
        zmq.PULL, connect=EVENT_IPC_ADDR)
    try:
        while True:
            data = await event_sock.read()
            event_name = data[0].decode('ascii')
            agent_id = data[1].decode('utf8')
            args = msgpack.unpackb(data[2])
            dispatcher.dispatch(event_name, agent_id, args)
    except asyncio.CancelledError:
        pass
    except:
        log.exception('unexpected error')
    finally:
        event_sock.close()


async def init(app):
    loop = asyncio.get_event_loop()
    dispatcher = EventDispatcher(app)
    app['event_dispatcher'] = dispatcher
    app['event_subscriber'] = loop.create_task(event_subscriber(dispatcher))


async def shutdown(app):
    app['event_subscriber'].cancel()
    await app['event_subscriber']



import asyncio
import inspect
import logging
import traceback
from typing import Mapping

from aiohttp import web
import graphene
from graphql.execution.executors.asyncio import AsyncioExecutor
from graphql.error.located_error import GraphQLLocatedError
import simplejson as json

from .exceptions import InvalidAPIParameters, BackendError
from .auth import auth_required
from ..manager.models.base import DataLoaderManager
from ..manager.models import (
    Agent,
    KeyPair, CreateKeyPair, ModifyKeyPair, DeleteKeyPair,
    ComputeSession, ComputeWorker, KernelStatus,
    VirtualFolder,
)

log = logging.getLogger('ai.backend.gateway.admin')


@auth_required
async def handle_gql(request: web.Request) -> web.Response:
    executor = request.app['admin.gql_executor']
    if request['is_admin']:
        schema = request.app['admin.gql_schema_admin']
    else:
        schema = request.app['admin.gql_schema_user']
    try:
        body = await request.json(loads=json.loads)
    except (asyncio.TimeoutError, json.decoder.JSONDecodeError):
        raise InvalidAPIParameters('Malformed request body.')
    try:
        assert 'query' in body, \
               'The request must have "query" JSON field.'
        assert isinstance(body['query'], str), \
               'The "query" field must be a JSON string.'
        if 'variables' in body:
            assert (body['variables'] is None or
                    isinstance(body['variables'], Mapping)), \
                   'The "variables" field must be an JSON object or null.'
        else:
            body['variables'] = None
    except AssertionError as e:
        raise InvalidAPIParameters(e.args[0])
    dlmanager = DataLoaderManager(request.app['dbpool'])
    result = schema.execute(
        body['query'], executor,
        variable_values=body['variables'],
        context_value={
            'dlmgr': dlmanager,
            'access_key': request['keypair']['access_key'],
            'dbpool': request.app['dbpool'],
            'redis_stat_pool': request.app['redis_stat_pool'],
        },
        return_promise=True)
    if inspect.isawaitable(result):
        result = await result
    if result.errors:
        has_internal_errors = False
        for e in result.errors:
            if isinstance(e, GraphQLLocatedError):
                exc_info = (type(e.original_error),
                            e.original_error,
                            e.original_error.__traceback__)
                tb_text = ''.join(traceback.format_exception(*exc_info))
                log.error(f'GraphQL located error:\n{tb_text}')
                request.app['sentry'].captureException(exc_info)
                has_internal_errors = True
        if has_internal_errors:
            raise BackendError(str(result.errors[0]))
        raise InvalidAPIParameters(str(result.errors[0]))
    else:
        return web.json_response(result.data, status=200, dumps=json.dumps)


class MutationForAdmin(graphene.ObjectType):
    create_keypair = CreateKeyPair.Field()
    modify_keypair = ModifyKeyPair.Field()
    delete_keypair = DeleteKeyPair.Field()


# Nothing yet!
# class MutationForUser(graphene.ObjectType):
#     pass


class QueryForAdmin(graphene.ObjectType):
    '''
    Available GraphQL queries for the admin privilege.
    It allows use of any access keys regardless of the one specified in the
    authorization header as well as querying the keypair information of all
    users.
    '''

    agent = graphene.Field(
        Agent,
        agent_id=graphene.String())

    agents = graphene.List(
        Agent,
        status=graphene.String())

    keypair = graphene.Field(
        KeyPair,
        access_key=graphene.String())

    keypairs = graphene.List(
        KeyPair,
        user_id=graphene.Int(required=True),
        is_active=graphene.Boolean())

    vfolders = graphene.List(
        VirtualFolder,
        access_key=graphene.String())

    compute_sessions = graphene.List(
        ComputeSession,
        access_key=graphene.String(),
        status=graphene.String())

    compute_workers = graphene.List(
        ComputeWorker,
        sess_id=graphene.String(required=True),
        status=graphene.String())

    @staticmethod
    async def resolve_agent(executor, info, agent_id):
        manager = info.context['dlmgr']
        loader = manager.get_loader('Agent', status=None)
        return await loader.load(agent_id)

    @staticmethod
    async def resolve_agents(executor, info, status=None):
        dbpool = info.context['dbpool']
        return await Agent.load_all(dbpool, status=status)

    @staticmethod
    async def resolve_keypair(executor, info, access_key=None):
        manager = info.context['dlmgr']
        if access_key is None:
            access_key = info.context['access_key']
        loader = manager.get_loader('KeyPair.by_ak')
        return await loader.load(access_key)

    @staticmethod
    async def resolve_keypairs(executor, info, user_id, is_active=None):
        manager = info.context['dlmgr']
        loader = manager.get_loader('KeyPair.by_uid', is_active=is_active)
        return await loader.load(user_id)

    @staticmethod
    async def resolve_vfolders(executor, info, access_key=None):
        manager = info.context['dlmgr']
        if access_key is None:
            access_key = info.context['access_key']
        loader = manager.get_loader('VirtualFolder')
        return await loader.load(access_key)

    @staticmethod
    async def resolve_compute_sessions(executor, info, access_key=None, status=None):
        manager = info.context['dlmgr']
        # TODO: make status a proper graphene.Enum type
        #       (https://github.com/graphql-python/graphene/issues/544)
        if access_key is None:
            access_key = info.context['access_key']
        if status is not None:
            status = KernelStatus[status]
        loader = manager.get_loader('ComputeSession', status=status)
        return await loader.load(access_key)

    @staticmethod
    async def resolve_compute_workers(executor, info, sess_id, status=None):
        manager = info.context['dlmgr']
        if status is not None:
            status = KernelStatus[status]
        loader = manager.get_loader('ComputeWorker', status=status)
        return await loader.load(sess_id)


class QueryForUser(graphene.ObjectType):
    '''
    Available GraphQL queries for the user priveilege.
    It only allows use of the access key specified in the authorization header.
    '''

    keypair = graphene.Field(lambda: KeyPair)

    vfolders = graphene.List(VirtualFolder)

    compute_sessions = graphene.List(
        ComputeSession,
        status=graphene.String())

    compute_workers = graphene.List(
        ComputeWorker,
        sess_id=graphene.String(required=True),
        status=graphene.String())

    @staticmethod
    async def resolve_keypair(executor, info):
        manager = info.context['dlmgr']
        access_key = info.context['access_key']
        loader = manager.get_loader('KeyPair.by_ak')
        return await loader.load(access_key)

    @staticmethod
    async def resolve_vfolders(executor, info):
        manager = info.context['dlmgr']
        access_key = info.context['access_key']
        loader = manager.get_loader('VirtualFolder')
        return await loader.load(access_key)

    @staticmethod
    async def resolve_compute_sessions(executor, info, status=None):
        manager = info.context['dlmgr']
        access_key = info.context['access_key']
        # TODO: make status a proper graphene.Enum type
        #       (https://github.com/graphql-python/graphene/issues/544)
        if status is not None:
            status = KernelStatus[status]
        loader = manager.get_loader('ComputeSession', status=status)
        return await loader.load(access_key)

    @staticmethod
    async def resolve_compute_workers(executor, info, sess_id, status=None):
        manager = info.context['dlmgr']
        access_key = info.context['access_key']
        if status is not None:
            status = KernelStatus[status]
        loader = manager.get_loader(
            'ComputeWorker', status=status, access_key=access_key)
        return await loader.load(sess_id)


async def init(app):
    loop = asyncio.get_event_loop()
    app.router.add_route('POST', r'/v{version:\d+}/admin/graphql', handle_gql)
    app['admin.gql_executor'] = AsyncioExecutor(loop=loop)
    app['admin.gql_schema_admin'] = graphene.Schema(
        query=QueryForAdmin,
        mutation=MutationForAdmin,
        auto_camelcase=False)
    app['admin.gql_schema_user'] = graphene.Schema(
        query=QueryForUser,
        mutation=None,
        auto_camelcase=False)


async def shutdown(app):
    pass


if __name__ == '__main__':
    # If executed as a main program, print all GraphQL schemas.
    # (graphene transforms our object model into a textual representation)
    # This is useful for writing documentation!
    admin_schema = graphene.Schema(
        query=QueryForAdmin,
        mutation=MutationForAdmin,
        auto_camelcase=False)
    user_schema = graphene.Schema(
        query=QueryForUser,
        mutation=None,
        auto_camelcase=False)
    print('======== Admin Schema ========')
    print(str(admin_schema))
    print('======== User Schema ========')
    print(str(user_schema))



'''
Kernel session management.
'''

import asyncio
import base64
from collections import defaultdict
from datetime import datetime
import functools
import logging
import secrets
import time
from urllib.parse import urlparse

import aiohttp
from aiohttp import web
import aiotools
import aiozmq
from aiozmq import create_zmq_stream as aiozmq_sock
from async_timeout import timeout as _timeout
from dateutil.tz import tzutc
import simplejson as json
import sqlalchemy as sa
from sqlalchemy.sql.expression import true, null
import zmq

from .exceptions import (ServiceUnavailable, InvalidAPIParameters, QuotaExceeded,
                         QueryNotImplemented, KernelNotFound,
                         BackendError)
from . import GatewayStatus
from .auth import auth_required
from .utils import catch_unexpected, method_placeholder
from ..manager.models import keypairs, kernels, AgentStatus, KernelStatus
from ..manager.registry import InstanceRegistry

log = logging.getLogger('ai.backend.gateway.kernel')

grace_events = []


def server_ready_required(handler):
    @functools.wraps(handler)
    async def wrapped(request):
        if request.app['status'] != GatewayStatus.RUNNING:
            raise ServiceUnavailable('Server not ready.')
        return (await handler(request))
    return wrapped


@auth_required
@server_ready_required
async def create(request):
    try:
        with _timeout(2):
            params = await request.json(loads=json.loads)
        log.info(f"GET_OR_CREATE (u:{request['keypair']['access_key']}, "
                 f"lang:{params['lang']}, token:{params['clientSessionToken']})")
        assert 8 <= len(params['clientSessionToken']) <= 80
    except (asyncio.TimeoutError, AssertionError,
            KeyError, json.decoder.JSONDecodeError) as e:
        log.warning(f'GET_OR_CREATE: invalid/missing parameters, {e!r}')
        raise InvalidAPIParameters
    resp = {}
    try:
        access_key = request['keypair']['access_key']
        concurrency_limit = request['keypair']['concurrency_limit']
        async with request.app['dbpool'].acquire() as conn, conn.begin():
            query = (sa.select([keypairs.c.concurrency_used], for_update=True)
                       .select_from(keypairs)
                       .where(keypairs.c.access_key == access_key))
            concurrency_used = await conn.scalar(query)
            log.debug(f'access_key: {access_key} '
                      f'({concurrency_used} / {concurrency_limit})')
            if concurrency_used >= concurrency_limit:
                raise QuotaExceeded
            if request['api_version'] == 1:
                limits = params.get('resourceLimits', None)
                mounts = None
            elif request['api_version'] in (2, 3):
                limits = params.get('limits', None)
                mounts = params.get('mounts', None)
            kernel, created = await request.app['registry'].get_or_create_kernel(
                params['clientSessionToken'],
                params['lang'], access_key,
                limits, mounts,
                conn=conn)
            resp['kernelId'] = str(kernel['sess_id'])
            if created:
                query = (sa.update(keypairs)
                           .values(concurrency_used=keypairs.c.concurrency_used + 1)
                           .where(keypairs.c.access_key == access_key))
                await conn.execute(query)
    except BackendError:
        log.exception('GET_OR_CREATE: exception')
        raise
    return web.json_response(resp, status=201, dumps=json.dumps)


@catch_unexpected(log)
async def update_instance_usage(app, inst_id):
    # In heartbeat timeouts, we do NOT clear Redis keys because
    # the timeout may be a transient one.
    kern_ids = await app['registry'].get_kernels_in_instance(inst_id)
    all_kernels = await app['registry'].get_kernels(kern_ids, allow_stale=True)
    affected_keys = [kern['access_key'] for kern in all_kernels if kern is not None]

    # TODO: enqueue termination event to streaming response queue

    per_key_counts = defaultdict(int)
    for ak in filter(lambda ak: ak is not None, affected_keys):
        per_key_counts[ak] += 1
    per_key_counts_str = ', '.join(f'{k}:{v}' for k, v in per_key_counts.items())
    log.info(f'-> cleaning {kern_ids!r}')
    log.info(f'-> per-key usage: {per_key_counts_str}')

    if not affected_keys:
        return

    async with app['dbpool'].acquire() as conn:
        log.debug(f'update_instance_usage({inst_id})')
        for kern in all_kernels:
            if kern is None:
                continue
            query = (sa.update(keypairs)
                       .values({
                           'concurrency_used': keypairs.c.concurrency_used -
                                               per_key_counts[kern['access_key']],  # noqa
                       })
                       .where(keypairs.c.access_key == kern['access_key']))
            await conn.execute(query)


@catch_unexpected(log)
async def kernel_terminated(app, agent_id, kernel_id, reason, kern_stat):
    try:
        kernel = await app['registry'].get_kernel(
            kernel_id, (kernels.c.role, kernels.c.status), allow_stale=True)
    except KernelNotFound:
        # Skip if missing
        return
    # Skip if restarting
    if kernel.status != KernelStatus.RESTARTING:
        await app['registry'].mark_kernel_terminated(kernel_id)
    if kernel.role == 'master':
        sess_id = kernel['sess_id']
        for handler in app['stream_pty_handlers'][sess_id].copy():
            handler.cancel()
            await handler
        if kernel.status != KernelStatus.RESTARTING:
            await app['registry'].mark_session_terminated(sess_id)


@catch_unexpected(log)
async def instance_started(app, agent_id):
    # TODO: make feedback to our auto-scaler
    await app['registry'].update_instance(agent_id, {
        'status': AgentStatus.ALIVE,
    })


@catch_unexpected(log)
async def instance_terminated(app, agent_id, reason):
    with app['shared_states'].lock:
        try:
            del app['shared_states'].agent_last_seen[agent_id]
        except KeyError:
            pass
    if reason == 'agent-lost':
        await app['registry'].mark_agent_terminated(agent_id, AgentStatus.LOST)
    elif reason == 'agent-restart':
        log.info(f'agent@{agent_id} restarting for maintenance.')
        await app['registry'].update_instance(agent_id, {
            'status': AgentStatus.RESTARTING,
        })
    else:
        # On normal instance termination, kernel_terminated events were already
        # triggered by the agent.
        await app['registry'].mark_agent_terminated(agent_id, AgentStatus.TERMINATED)


@catch_unexpected(log)
async def instance_heartbeat(app, agent_id, agent_info):
    with app['shared_states'].lock:
        app['shared_states'].agent_last_seen[agent_id] = time.monotonic()
    await app['registry'].handle_heartbeat(agent_id, agent_info)


@catch_unexpected(log)
async def check_agent_lost(app, interval):
    try:
        now = time.monotonic()
        with app['shared_states'].lock:
            copied = app['shared_states'].agent_last_seen.copy()
        for agent_id, prev in copied.items():
            if now - prev >= app.config.heartbeat_timeout:
                # TODO: change this to "send_event" (actual zeromq push)
                #       for non-duplicate events
                app['event_dispatcher'].dispatch('instance_terminated',
                                                 agent_id, ('agent-lost', ))
    except asyncio.CancelledError:
        pass


# NOTE: This event is ignored during the grace period.
@catch_unexpected(log)
async def instance_stats(app, agent_id, kern_stats):
    await app['registry'].handle_stats(agent_id, kern_stats)


async def datadog_update(app):
    with app['datadog'].statsd as statsd:

        statsd.gauge('ai.backend.gateway.coroutines', len(asyncio.Task.all_tasks()))

        all_inst_ids = [
            inst_id async for inst_id
            in app['registry'].enumerate_instances()]
        statsd.gauge('ai.backend.gateway.agent_instances', len(all_inst_ids))

        async with app['dbpool'].acquire() as conn:
            query = (sa.select([sa.func.sum(keypairs.c.concurrency_used)])
                       .select_from(keypairs))
            n = await conn.scalar(query)
            statsd.gauge('ai.backend.gateway.active_kernels', n)

            subquery = (sa.select([sa.func.count()])
                          .select_from(keypairs)
                          .where(keypairs.c.is_active == true())
                          .group_by(keypairs.c.user_id))
            query = sa.select([sa.func.count()]).select_from(subquery.alias())
            n = await conn.scalar(query)
            statsd.gauge('ai.backend.users.has_active_key', n)

            subquery = subquery.where(keypairs.c.last_used != null())
            query = sa.select([sa.func.count()]).select_from(subquery.alias())
            n = await conn.scalar(query)
            statsd.gauge('ai.backend.users.has_used_key', n)

            '''
            query = sa.select([sa.func.count()]).select_from(usage)
            n = await conn.scalar(query)
            statsd.gauge('ai.backend.gateway.accum_kernels', n)
            '''


async def datadog_update_timer(app):
    if app['datadog'] is None:
        return
    while True:
        try:
            await datadog_update(app)
        except asyncio.CancelledError:
            break
        except:
            app['sentry'].captureException()
            log.exception('datadog_update unexpected error')
        try:
            await asyncio.sleep(5)
        except asyncio.CancelledError:
            break


@auth_required
@server_ready_required
async def destroy(request):
    sess_id = request.match_info['sess_id']
    log.info(f"DESTROY (u:{request['keypair']['access_key']}, k:{sess_id})")
    try:
        await request.app['registry'].destroy_kernel(sess_id)
    except BackendError:
        log.exception('DESTROY: exception')
        raise
    return web.Response(status=204)


@auth_required
@server_ready_required
async def get_info(request):
    resp = {}
    sess_id = request.match_info['sess_id']
    log.info(f"GETINFO (u:{request['keypair']['access_key']}, k:{sess_id})")
    try:
        await request.app['registry'].increment_session_usage(sess_id)
        kern = await request.app['registry'].get_kernel_session(sess_id, field='*')
        resp['lang'] = kern.lang
        age = datetime.now(tzutc()) - kern.created_at
        resp['age'] = age.total_seconds() * 1000
        # Resource limits collected from agent heartbeats
        # TODO: factor out policy/image info as a common repository
        resp['queryTimeout']  = -1  # deprecated
        resp['idleTimeout']   = -1  # deprecated
        resp['memoryLimit']   = kern.mem_max_bytes >> 10  # KiB
        resp['maxCpuCredit']  = -1  # deprecated
        # Stats collected from agent heartbeats
        resp['numQueriesExecuted'] = kern.num_queries
        resp['idle']          = -1  # deprecated
        resp['memoryUsed']    = kern.mem_cur_bytes >> 10  # KiB
        resp['cpuCreditUsed'] = kern.cpu_used
        log.info(f'information retrieved: {resp!r}')
    except BackendError:
        log.exception('GETINFO: exception')
        raise
    return web.json_response(resp, status=200, dumps=json.dumps)


@auth_required
@server_ready_required
async def restart(request):
    sess_id = request.match_info['sess_id']
    log.info(f"RESTART (u:{request['keypair']['access_key']}, k:{sess_id})")
    try:
        await request.app['registry'].increment_session_usage(sess_id)
        await request.app['registry'].restart_kernel(sess_id)
        for sock in request.app['stream_stdin_socks'][sess_id]:
            sock.close()
    except BackendError:
        log.exception('RESTART: exception')
        raise
    except:
        request.app['sentry'].captureException()
        log.exception('RESTART: unexpected error')
        raise web.HTTPInternalServerError
    return web.Response(status=204)


@auth_required
@server_ready_required
async def execute(request):
    resp = {}
    sess_id = request.match_info['sess_id']
    try:
        with _timeout(2):
            params = await request.json(loads=json.loads)
        log.info(f"EXECUTE(u:{request['keypair']['access_key']}, k:{sess_id})")
    except (asyncio.TimeoutError, json.decoder.JSONDecodeError):
        log.warning('EXECUTE: invalid/missing parameters')
        raise InvalidAPIParameters
    try:
        await request.app['registry'].increment_session_usage(sess_id)
        api_version = request['api_version']
        if api_version == 1:
            mode = 'query'
            code = params['code']
            run_id = params.get('runId', secrets.token_hex(8))
            opts = {}
        elif api_version in (2, 3):
            mode = params['mode']
            code = params.get('code', '')
            run_id = params.get('runId', secrets.token_hex(8))
            mode = params['mode']
            assert mode in ('query', 'batch', 'complete')
            opts = params.get('options', None) or {}
        if mode == 'complete':
            # For legacy
            resp['result'] = await request.app['registry'].get_completions(
                sess_id, 'query', code, opts)
        else:
            resp['result'] = await request.app['registry'].execute(
                sess_id, api_version, run_id, mode, code, opts)
    except AssertionError:
        log.warning('EXECUTE: invalid/missing parameters')
        raise InvalidAPIParameters
    except BackendError:
        log.exception('EXECUTE: exception')
        raise
    return web.json_response(resp, status=200, dumps=json.dumps)


@auth_required
@server_ready_required
async def interrupt(request):
    sess_id = request.match_info['sess_id']
    log.info(f"INTERRUPT(u:{request['keypair']['access_key']}, k:{sess_id})")
    try:
        await request.app['registry'].increment_session_usage(sess_id)
        await request.app['registry'].interrupt_kernel(sess_id)
    except BackendError:
        log.exception('INTERRUPT: exception')
        raise
    return web.Response(status=204)


@auth_required
@server_ready_required
async def complete(request):
    resp = {}
    sess_id = request.match_info['sess_id']
    try:
        with _timeout(2):
            params = await request.json(loads=json.loads)
        log.info(f"COMPLETE(u:{request['keypair']['access_key']}, k:{sess_id})")
    except (asyncio.TimeoutError, json.decoder.JSONDecodeError):
        log.warning('COMPLETE: invalid/missing parameters')
        raise InvalidAPIParameters
    try:
        code = params.get('code', '')
        opts = params.get('options', None) or {}
        await request.app['registry'].increment_session_usage(sess_id)
        resp['result'] = await request.app['registry'].get_completions(
            sess_id, 'query', code, opts)
    except AssertionError:
        log.warning('COMPLETE: invalid/missing parameters')
        raise InvalidAPIParameters
    except BackendError:
        log.exception('COMPLETE: exception')
        raise
    return web.json_response(resp, status=200, dumps=json.dumps)


@auth_required
@server_ready_required
async def upload_files(request):
    loop = asyncio.get_event_loop()
    reader = await request.multipart()
    sess_id = request.match_info['sess_id']
    try:
        await request.app['registry'].increment_session_usage(sess_id)
        file_count = 0
        upload_tasks = []
        while True:
            if file_count == 20:
                raise InvalidAPIParameters('Too many files')
            file = await reader.next()
            if file is None:
                break
            file_count += 1
            # This API handles only small files, so let's read it at once.
            chunk = await file.read_chunk(size=1048576)
            if not file.at_eof():
                raise InvalidAPIParameters('Too large file')
            data = file.decode(chunk)
            log.debug(f'received file: {file.filename} ({len(data):,} bytes)')
            t = loop.create_task(
                request.app['registry'].upload_file(sess_id, file.filename, data))
            upload_tasks.append(t)
        await asyncio.gather(*upload_tasks)
    except BackendError:
        log.exception('UPLOAD_FILES: exception')
        raise
    return web.Response(status=204)


@server_ready_required
async def stream_pty(request):
    app = request.app
    sess_id = request.match_info['sess_id']
    try:
        kernel = await app['registry'].get_kernel_session(sess_id, field=(
            kernels.c.stdin_port, kernels.c.stdout_port
        ))
    except KernelNotFound:
        raise

    await app['registry'].increment_session_usage(sess_id)

    # Upgrade connection to WebSocket.
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    app['stream_pty_handlers'][sess_id].add(asyncio.Task.current_task())

    async def connect_streams(kernel):
        kernel_ip = urlparse(kernel.agent_addr).hostname
        stdin_addr = f'tcp://{kernel_ip}:{kernel.stdin_port}'
        log.debug(f'stream_pty({sess_id}): stdin: {stdin_addr}')
        stdin_sock = await aiozmq_sock(zmq.PUB, connect=stdin_addr)
        stdin_sock.transport.setsockopt(zmq.LINGER, 100)
        stdout_addr = f'tcp://{kernel_ip}:{kernel.stdout_port}'
        log.debug(f'stream_pty({sess_id}): stdout: {stdout_addr}')
        stdout_sock = await aiozmq_sock(zmq.SUB, connect=stdout_addr)
        stdout_sock.transport.setsockopt(zmq.LINGER, 100)
        stdout_sock.transport.subscribe(b'')
        return stdin_sock, stdout_sock

    # Wrap sockets in a list so that below coroutines can share reference changes.
    socks = list(await connect_streams(kernel))
    app['stream_stdin_socks'][sess_id].add(socks[0])
    stream_sync = asyncio.Event()

    async def stream_stdin():
        nonlocal socks
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    data = json.loads(msg.data)
                    if data['type'] == 'stdin':
                        raw_data = base64.b64decode(data['chars'].encode('ascii'))
                        try:
                            socks[0].write([raw_data])
                        except (AttributeError, aiozmq.ZmqStreamClosed):
                            # AttributeError occurs when stdin_sock._transport
                            # is None because it's already closed somewhere
                            # else.
                            app['stream_stdin_socks'][sess_id].remove(socks[0])
                            socks[1].close()
                            kernel = \
                                await app['registry'].get_kernel_session(sess_id)
                            stdin_sock, stdout_sock = await connect_streams(kernel)
                            socks[0] = stdin_sock
                            socks[1] = stdout_sock
                            app['stream_stdin_socks'][sess_id].add(socks[0])
                            socks[0].write([raw_data])
                            log.debug(f'stream_stdin({sess_id}): zmq stream reset')
                            stream_sync.set()
                            continue
                    else:
                        await app['registry'].increment_session_usage(sess_id)
                        api_version = 2
                        run_id = secrets.token_hex(8)
                        if data['type'] == 'resize':
                            code = f"%resize {data['rows']} {data['cols']}"
                            await app['registry'].execute(
                                sess_id, api_version, run_id, 'query', code, {})
                        elif data['type'] == 'ping':
                            await app['registry'].execute(
                                sess_id, api_version, run_id, 'query', '%ping', {})
                        elif data['type'] == 'restart':
                            # Close existing zmq sockets and let stream
                            # handlers get a new one with changed stdin/stdout
                            # ports.
                            if not socks[0].at_closing():
                                await app['registry'].restart_kernel(sess_id)
                                socks[0].close()
                            else:
                                log.warning(f'stream_stdin({sess_id}): '
                                            'duplicate kernel restart request; '
                                            'ignoring it.')
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    log.warning(f'stream_stdin({sess_id}): '
                                f'connection closed ({ws.exception()})')
        except asyncio.CancelledError:
            # Agent or kernel is terminated.
            pass
        except:
            app['sentry'].captureException()
            log.exception(f'stream_stdin({sess_id}): unexpected error')
        finally:
            log.debug(f'stream_stdin({sess_id}): terminated')
            if not socks[0].at_closing():
                socks[0].close()

    async def stream_stdout():
        nonlocal socks
        log.debug(f'stream_stdout({sess_id}): started')
        try:
            while True:
                try:
                    data = await socks[1].read()
                except aiozmq.ZmqStreamClosed:
                    await stream_sync.wait()
                    stream_sync.clear()
                    log.debug(f'stream_stdout({sess_id}): zmq stream reset')
                    continue
                if ws.closed:
                    break
                ws.send_str(json.dumps({
                    'type': 'out',
                    'data': base64.b64encode(data[0]).decode('ascii'),
                }, ensure_ascii=False))
        except asyncio.CancelledError:
            pass
        except:
            app['sentry'].captureException()
            log.exception(f'stream_stdout({sess_id}): unexpected error')
        finally:
            log.debug(f'stream_stdout({sess_id}): terminated')
            socks[1].close()

    # According to aiohttp docs, reading ws must be done inside this task.
    # We execute the stdout handler as another task.
    try:
        stdout_task = asyncio.ensure_future(stream_stdout())
        await stream_stdin()
    except:
        app['sentry'].captureException()
        log.exception(f'stream_pty({sess_id}): unexpected error')
    finally:
        app['stream_pty_handlers'][sess_id].remove(asyncio.Task.current_task())
        app['stream_stdin_socks'][sess_id].remove(socks[0])
        stdout_task.cancel()
        await stdout_task
    return ws


@auth_required
async def not_impl_stub(request):
    raise QueryNotImplemented


async def init(app):
    rt = app.router.add_route
    rt('POST',   r'/v{version:\d+}/kernel/create', create)
    rt('GET',    r'/v{version:\d+}/kernel/{sess_id}', get_info)

    rt('PATCH',  r'/v{version:\d+}/kernel/{sess_id}', restart)
    rt('DELETE', r'/v{version:\d+}/kernel/{sess_id}', destroy)
    rt('POST',   r'/v{version:\d+}/kernel/{sess_id}', execute)
    rt('POST',   r'/v{version:\d+}/kernel/{sess_id}/interrupt', interrupt)
    rt('POST',   r'/v{version:\d+}/kernel/{sess_id}/complete', complete)
    rt('GET',    r'/v{version:\d+}/stream/kernel/{sess_id}/pty', stream_pty)
    rt('GET',    r'/v{version:\d+}/stream/kernel/{sess_id}/events', not_impl_stub)
    rt('POST',   r'/v{version:\d+}/kernel/{sess_id}/upload', upload_files)
    rt('POST',   r'/v{version:\d+}/folder/create', not_impl_stub)
    rt('GET',    r'/v{version:\d+}/folder/{folder_id}', not_impl_stub)
    rt('POST',   r'/v{version:\d+}/folder/{folder_id}', method_placeholder('DELETE'))
    rt('DELETE', r'/v{version:\d+}/folder/{folder_id}', not_impl_stub)

    app['event_dispatcher'].add_handler('kernel_terminated', kernel_terminated)
    app['event_dispatcher'].add_handler('instance_started', instance_started)
    app['event_dispatcher'].add_handler('instance_terminated', instance_terminated)
    app['event_dispatcher'].add_handler('instance_heartbeat', instance_heartbeat)
    app['event_dispatcher'].add_handler('instance_stats', instance_stats)

    app['stream_pty_handlers'] = defaultdict(set)
    app['stream_stdin_socks'] = defaultdict(set)

    app['registry'] = InstanceRegistry(app['dbpool'], app['redis_stat_pool'])
    await app['registry'].init()

    # Scan ALIVE agents
    if app['pidx'] == 0:
        log.debug(f'initializing agent status checker at proc:{app["pidx"]}')
        now = time.monotonic()
        async for inst in app['registry'].enumerate_instances():
            app['shared_states'].agent_last_seen[inst.id] = now
        app['agent_lost_checker'] = aiotools.create_timer(
            functools.partial(check_agent_lost, app), 1.0)

    app['shared_states'].barrier.wait()
    app['status'] = GatewayStatus.RUNNING


async def shutdown(app):
    if app['pidx'] == 0:
        app['agent_lost_checker'].cancel()
        await app['agent_lost_checker']

    checked_tasks = ('kernel_agent_event_collector', 'kernel_ddtimer')
    for tname in checked_tasks:
        t = app.get(tname, None)
        if t and not t.done():
            t.cancel()
            await t
    for per_kernel_handlers in app['stream_pty_handlers'].values():
        for handler in per_kernel_handlers.copy():
            handler.cancel()
            await handler
    await app['registry'].terminate()
